# Copyright 2020 William Ro. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ====-========================================================================-
"""City hall"""

from ..spqr.arguments import Arguments
from ..spqr.censor import check_type
from ..console.console import console

from . import city
from . import library


class Hall(object):

  def __init__(self, city_, **funcs):
    self.city = check_type(city_, city.City)
    self.funcs = funcs
    # Register build-in functions
    library.register(city_, self.funcs)


  def handle(self, cmd):
    # Parse command
    a = Arguments.parse(cmd)
    if a.func_name not in self.funcs:
      console.show_status(
        "Unknown command '{}'".format(a.func_name), prompt='!!', color='red')
      return False
    return self.funcs[a.func_name](*a.arg_list, **a.arg_dict)




# Copyright 2020 William Ro. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==-========================================================================-==
from collections import OrderedDict



class Nomear(object):
  """This class is designed to provide a `pocket` dictionary for instances
  of classes inherited from it. If you are curious about its name, append
  `od` and reverse. One of the cool things about this class is that users
  do not need to call its constructor/initializer.

  This class follows the coding philosophy that some fields of a certain
  class do not need to be explicitly appear in constructor. Sometimes put
  them into a pocket is a better way.
  """

  _LOCAL_POCKET_KEY = '_NOMEAR_LOCAL_POCKET'

  cloud = OrderedDict()


  @property
  def _cloud_pocket(self) -> OrderedDict:
    # Register self in Nomear cloud if necessary
    if self not in self.cloud: self.cloud[self] = OrderedDict()
    return self.cloud[self]


  @property
  def _local_pocket(self) -> OrderedDict:
    # Initialize local pocket if necessary
    if not hasattr(self, self._LOCAL_POCKET_KEY):
      setattr(self, self._LOCAL_POCKET_KEY, OrderedDict())
    return getattr(self, self._LOCAL_POCKET_KEY)


  @property
  def _pocket(self) -> OrderedDict:
    """Gather all stuff in local and cloud pocket and return"""
    p = self._cloud_pocket.copy()
    p.update(self._local_pocket)
    return p


  def in_pocket(self, key): return key in self._pocket


  def localize(self, key, exclusive=False, key_should_exist=False):
    if key not in self._cloud_pocket:
      if key_should_exist: raise KeyError(
        "!! `{}` not found in {}'s cloud pocket.".format(key, self))
      else: return None

    return self.put_into_pocket(
      key, self._cloud_pocket[key], exclusive, local=True)


  def get_from_pocket(
      self, key: str, default=None, initializer=None, local=False,
      key_should_exist=False, put_back=True):
    if not self.in_pocket(key):
      if callable(initializer):
        return self.put_into_pocket(key, initializer(), local=local)
      elif key_should_exist: raise KeyError(
        "!! `{}` not found in {}'s pockets.".format(key, self))
      else: return default
    if put_back: return self._pocket[key]
    # Pop if not put back
    if key in self._cloud_pocket: return self._cloud_pocket.pop(key)
    return self._local_pocket.pop(key)


  def put_into_pocket(self, key: str, thing, exclusive=True, local=False):
    pocket = self._local_pocket if local else self._cloud_pocket
    if key in pocket and exclusive: raise KeyError(
      "`{}` already exists in {}'s {} pocket.".format(
        key, self, 'local' if local else 'cloud'))
    pocket[key] = thing
    return thing


  def replace_stuff(self, key: str, val, local=False):
    pocket = self._local_pocket if local else self._cloud_pocket
    assert key in pocket
    pocket[key] = val


  def release(self):
    if self in self.cloud: self.cloud.pop(self)


  def __getitem__(self, item):
    return self.get_from_pocket(item, key_should_exist=True)


  def __setitem__(self, key, value):
    self.put_into_pocket(key, value)


  @staticmethod
  def property(local=False, key=None):
    def _decorator(func):
      _key = str(func).split()[1].split('.')[1] if key is None else key
      @property
      def _func(self):
        assert isinstance(self, Nomear)
        return self.get_from_pocket(
          _key, initializer=lambda: func(self), local=local)
      return _func
    return _decorator



# Copyright 2020 William Ro. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ====-==============================================================-==========
"""This module provides methods related to types of method arguments.
"""
from ..spqr import atticus


def check_type(input_, type_=None, inner_type=None, nullable=False,
               auto_conversion=True):
  """Check the type of the given inputs. This method can also be used as a
  parser for converting strings to numbers.

  When type_ is a groups of more than 1 types, auto conversion will not be
  performed even when auto_conversion is True.

  When input_ is a large group, calling this method frequently may bring
  time overhead which is not neglectable since it involves process to build
  a list of the same scale as the input group.

  Examples:
  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  >>> check_type(31, int)
  31

  >>> check_type({'adh', 93, 19}, inner_type=str)
  {None, 93, 19}

  >>> check_type([None, 12, 19.0], tuple, inner_type=int, nullable=True)
  (None, 12, 19),
  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  :param input_: input to be checked
  :param type_: a type or tuple/list/set of types, None by default.
                When this arg is None, inner_type must be specified and
                input_ is assumed to be in group types.
  :param inner_type: if specified, arg type_ must be None or one of group_types
  :param nullable: case 1: inner_type is None: whether input_ can be None
                   case 2: inner_type is not None: whether value in input_
                           can be None
  :param auto_conversion: whether to perform automatic type conversion.
                          Works only when arg 2 is a type.
  :return: input_, may be converted
  """
  group_types = (tuple, list, set)

  # Sanity check
  if all([type_ is None, inner_type is None]): raise ValueError(
      '!! check_type() arg 2 and 3 can not be None at the same time')

  # Check nullable (logically this line is not necessary)
  if all([nullable, input_ is None, inner_type is None]): return None

  # Make sure type_/inner_type is legal for isinstance() arg 2.
  # After this code block, type_/inner_type will be a type or a tuple of at
  # least 2 types
  def _check_type_type(_type, arg_ind):
    assert arg_ind in (2, 3)
    if _type is None: return None
    if isinstance(_type, group_types):
      type_is_legal = all([isinstance(t, type) for t in _type])
      _type = tuple(_type) if len(_type) > 1 else _type[0]
    else: type_is_legal = isinstance(_type, type)
    if not type_is_legal: raise TypeError(
      '!! check_type() arg {} must be a type or tuple/list/set of types'.format(
        arg_ind))
    return _type

  type_ = _check_type_type(type_, 2)
  inner_type = _check_type_type(inner_type, 3)

  # Define some inner methods for checking input_
  def _safe_convert(src, tgt_type):
    tgt = tgt_type(src)
    if all([isinstance(src, float), isinstance(tgt, int), src != tgt]):
      raise ValueError('!! Failed to convert {} to integer.'.format(src))
    return tgt

  def _raise_error(val, name, _type, convert=False):
    msg = '!! {0} (value: {1}) does not match{3} the required type {2}.'.format(
      name, val, _type, ' and can not be converted to' if convert else '')
    raise TypeError(msg)

  def _check_type(inp, _type, name='check_type() arg 1'):
    if not isinstance(inp, _type):
      if nullable and inp is None: return None
      if isinstance(_type, type) and auto_conversion:
        try: return _safe_convert(inp, _type)
        except: _raise_error(inp, name, _type, convert=True)
      _raise_error(inp, name, _type, convert=False)
    return inp

  # Check input_
  if type_ is None: type_ = group_types
  input_ = _check_type(input_, type_)
  if inner_type is None: return input_
  # Make sure input is a group
  assert inner_type is not None and isinstance(input_, group_types)
  return type(input_)(
    [_check_type(inp, inner_type, 'The {} element of check_type() arg 1'.format(
      atticus.ordinal(i + 1))) for i, inp in enumerate(input_)])


if __name__ == '__main__':
  assert all([
    check_type(31, int) == 31,
    check_type({'adh', 93, 19}, inner_type=str) == {'adh', '93', '19'},
    check_type([None, 12, 19.0], tuple,
               inner_type=int, nullable=True) == (None, 12, 19),
  ])



# Copyright 2020 William Ro. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ====-========================================================================-
"""Build-in functions for a city is maintained in this package"""

from .scroll import Scroll

from .basic import ExitHall
from .commu import SendMsg

from .. import city


CONTENT = (
  ExitHall,
  SendMsg,
)


def register(city_, func_dict: dict):
  assert isinstance(city_, city.City)
  for s in CONTENT:
    assert issubclass(s, Scroll)
    s(city_).register(func_dict)





# Copyright 2020 William Ro. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==-==========================================================================-
"""Titus Pomponius Atticus was a Roman editor who loved letters.
"""
from ..spqr import censor


def ordinal(n):
  """Convert a non-negative integer to its ordinal representation.

  :param n: a non-negative integer
  :return: the corresponding ordinal representation of input n
  """
  # Make sure n is a non-negative integer
  censor.check_type(n, int)
  # The terse solution below is taken from
  # http://golf.shinh.org/reveal.rb?1st%202nd%203rd%204th/xsot_1456205408&py
  return '{}{}'.format(n, 'tsnrhtdd'[n%5*(n%100^15>4>n%10)::4])


if __name__ == '__main__':
  for n in range(150): print(ordinal(n))




# Copyright 2020 William Ro. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ====-==============================================================-==========
"""A class that provides utilities to show information in terminal.
"""
import logging
import time
import warnings
from . import printer
from functools import wraps


def auto_clear(func):
  @wraps(func)
  def wrapper(*args, **kwargs):
    # Get the instance of Console class from args
    self = args[0]
    assert isinstance(self, Console)
    # Clear line if last called function is print_progress
    # Using 'is' instead of '==' does not work
    if self._last_func_called == self.print_progress:
      self.clear_line()
    self._last_func_called = func
    # Print
    return func(*args, **kwargs)
  return wrapper


class Console(object):

  __VERSION__ = (1, 0, 1)

  DEFAULT_TITLE = 'main'
  DEFAULT_PROMPT = '>>'
  INFO_PROMPT = '::'
  SUPPLEMENT_PROMPT = '..'
  WARNING_PROMPT = '!!'
  TEXT_WIDTH = 79

  def __init__(self, buffer_size=0, fancy_text=True):
    """Initiate a console.

    :param buffer_size: buffer size
    :param fancy_text: whether to allow fancy text. Notice that if this value
                       is set to False, all consoles will not be able to produce
                       fancy text.
    """
    # Public variables
    self.buffer_size = buffer_size
    assert isinstance(buffer_size, int) and buffer_size >= 0
    # Turn off fancy text forever if fancy_text is set to False
    if not fancy_text: self.disable_fancy_text()
    # Private variables
    self._buffer = []
    self._title = None
    self._last_func_called = None
    self._tic = None

  # region: Properties

  @property
  def buffer(self):
    return self._buffer

  @property
  def buffer_string(self):
    return '\n'.join(self._buffer)

  # endregion: Properties

  # region: Private Methods

  def _add_to_buffer(self, line):
    assert isinstance(line, str)
    if self.buffer_size == 0: return
    assert self.buffer_size > 0
    self._buffer.append(line)
    if len(self._buffer) > self.buffer_size: self._buffer.pop(0)

  # endregion: Private Methods
  
  # region: Public Methods

  @auto_clear
  def write_line(self, text, color=None, highlight=None, attributes=None,
                 buffer=True, **kwargs):
    """Write a line with fancy style.

    During parsing, changing the leading character is not supported currently.

    :param text: text to be written
    :param color: should be in
          {red, green, yellow, blue, magenta, cyan, white}
    :param highlight: should be in
          {on_red, on_green, on_yellow, on_blue, on_magenta, on_cyan, on_white}
    :param attributes: should be a subset of
          {bold, dark, underline, blink, reverse, concealed}
    :param buffer: whether to add text to buffer
    :param kwargs: additional keyword arguments for python print function
    :return: raw text
    """
    text = printer.write_line(text, color, highlight, attributes, **kwargs)
    # Add to buffer if necessary
    if buffer: self._add_to_buffer(text)
    return text


  def write(self, text, color=None, highlight=None, attributes=None):
    """Write the given text with fancy style, as in write_line method.

    During parsing, changing the leading character is not supported currently.
    This method will neither return anything nor add content into buffer.

    :param text: text to be written
    """
    printer.write(text, color, highlight, attributes)


  def split(self, splitter='-', text_width=None, color=None, highlight=None,
            attributes=None):
    """Split terminal with fancy (if style is specified) splitter.

    Example:
      console.split('#{-}{red}#{-}{yellow}#{-}{blue}')
    """
    text_width = text_width or self.TEXT_WIDTH
    # Repeat number should be calculated using raw text
    ren, raw = printer.parse(splitter)
    num = int(text_width / len(raw))
    self.write_line(num * ren, color, highlight, attributes)


  def start(self, title=None, text_width=None, color='grey'):
    """Indicate the start of a program."""
    title = title or self.DEFAULT_TITLE
    text_width = text_width or self.TEXT_WIDTH
    self.write_line('-> Start of {}'.format(title), color)
    self.split(text_width=text_width, color=color)
    self._title = title


  def end(self, text_width=None, color='grey'):
    """Indicate the end of a program."""
    text_width = text_width or self.TEXT_WIDTH
    self.split(text_width=text_width, color=color)
    self.write_line('|> End of {}'.format(self._title), color)


  def section(self, section_title):
    """Indicate the begin of a section."""
    self.split()
    self.write_line(':: {}'.format(section_title))
    self.split()


  @auto_clear
  def show_status(self, text, color=None, highlight=None, attributes=None,
                 prompt=None, buffer=True, **kwargs):
    """Show text following the specified prompt.

    The text will be displayed in a fancy style if corresponding configurations
    are provided, as in self.write_line method.

    :param text: text to be written.
    :param prompt: The leading symbol, usually '>>' by default.
    :param buffer: Whether to add text to buffer
    """
    # Use default prompt symbol if not provided
    if prompt is None: prompt = self.DEFAULT_PROMPT
    # The type of 'text' should not be restricted
    assert isinstance(prompt, str)
    prompt_text = '{} {}'.format(prompt, text)
    # Print using write_line
    return self.write_line(
      prompt_text, color, highlight, attributes, buffer=buffer, **kwargs)


  def show_info(self, text, color=None, highlight=None, attributes=None):
    """Show information using self.show_status"""
    return self.show_status(
      text, color, highlight, attributes, self.INFO_PROMPT)


  def supplement(self, text, color=None, highlight=None, attributes=None,
                 level=1):
    """Show supplement using self.show_status"""
    assert isinstance(level, int) and level > 0
    return self.show_status(
      text, color, highlight, attributes, self.SUPPLEMENT_PROMPT * level)


  def warning(self, text, color='red', highlight=None, attributes=None):
    """Show supplement using self.show_status"""
    return self.show_status(
      text, color, highlight, attributes, self.WARNING_PROMPT)


  def print_progress(self, index=None, total=None, progress=None):
    """Show progress bar using printer.print_progress.

    :param index: positive scalar, indicating current working progress
    :param total: positive scalar, indicating the scale of total work
    :param progress: if provided, 'index' and 'total' will be ignored.
    """
    if index == 0: self._tic = time.time()
    printer.print_progress(index, total, self._tic, progress)
    # This method does not need to be decorated due to the line below
    self._last_func_called = self.print_progress

  # endregion: Public Methods

  # region: Static Methods
  
  @staticmethod
  def fancify(text, *args):
    """Wrap raw text in the format that can be parsed by printer (BETA)"""
    result = '#{' + text + '}'
    for arg in args:
      assert isinstance(arg, str)
      result += '{' + arg + '}'
    return result


  @staticmethod
  def clear_line():
    """Clear a line in which current cursor is positioned."""
    printer.clear_line(Console.TEXT_WIDTH)


  @staticmethod
  def disable_fancy_text():
    """Disable fancy text. Notice that this will disable fancy text for all
       instances of Console.
    """
    printer.fancy_text = False


  @staticmethod
  def disable_future_warnings():
    """Suppress the annoying future warnings for good."""
    warnings.simplefilter(action='ignore', category=FutureWarning)


  @staticmethod
  def disable_logging(pkg_name):
    """Suppress the annoying logging information in terminal.
    :param pkg_name: Name of the package producing the unwanted log.
                     e.g., 'tensorflow'
    """
    assert isinstance(pkg_name, str)
    logging.getLogger(pkg_name).disabled = True

  # endregion: Static Methods






# Copyright 2020 William Ro. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ====-=====================================================================-==
"""This module provides methods for manipulating files"""
from ..console.console import console
from fnmatch import fnmatch

import os
import re
import zipfile


def walk(root_path, type_filter=None, pattern=None, ignored_patterns=(),
         re_pattern=None, return_basename=False, recursive=False,
         ignore_hidden_directories=True, include_folder_name=False) -> list:
  """Traverse through all files/folders under given `root_path`

  :param root_path: directory to be traversed
  :param type_filter: specify what type of path should be returned, can be
                      (1) None (default), indicating return all paths
                      (2) `file` for filtering out file paths
                      (3) `folder` or `dir` for filtering out directories
  :param pattern: patterns to be included using fnmatch
  :param ignored_patterns: patterns to be ignored using fnmatch
  :param re_pattern: patterns to be included using regular expression
  :param return_basename: whether to return the base name
  :param recursive: whether to traverse recursively.
                    If set to True, `return_basename` must be False
  :param ignore_hidden_directories: whether to ignore hidden directories
  :param include_folder_name: whether to include folder name when `type_filter`
                              is `file`. Designed for `tree` method
  :return: a list of paths
  """
  # Sanity check
  if not os.path.exists(root_path): raise FileNotFoundError(
    '!! Directory `{}` does not exist.'.format(root_path))

  # Get all paths under root_path
  paths = [os.path.join(root_path, p).replace('\\', '/')
           for p in os.listdir(root_path)]

  # Define filter function
  def filter_out(f, filter_full_path=False):
    if not callable(f): return paths
    _f = f if filter_full_path else lambda p: f(os.path.basename(p))
    return list(filter(_f, paths))

  # Filter path
  if type_filter in ('file', os.path.isfile):
    type_filter = os.path.isfile
  elif type_filter in ('folder', 'dir', os.path.isdir):
    type_filter = os.path.isdir
  elif type_filter is not None: raise ValueError(
    '!! `type_filter` should be one of (`file`, `folder`, `dir`)')
  paths = filter_out(type_filter, filter_full_path=True)

  # Include pattern using fnmatch if necessary
  if pattern is not None:
    paths = filter_out(lambda p: fnmatch(p, pattern))

  # Ignore patterns using fnmatch if necessary
  if isinstance(ignored_patterns, str): ignored_patterns = ignored_patterns,
  assert isinstance(ignored_patterns, (tuple, list))
  if ignore_hidden_directories:
    ignored_patterns = list(ignored_patterns) + ['.*']
  if len(ignored_patterns) > 0:
    paths = filter_out(
      lambda p: not any([fnmatch(p, ptn) for ptn in ignored_patterns]))

  # Filter pattern using regular expression if necessary
  if re_pattern is not None:
    paths = filter_out(lambda p: re.match(re_pattern, p) is not None)

  # Return if required
  if return_basename:
    if recursive: raise AssertionError(
      '!! Can not return base name when traversing recursively')
    return [os.path.basename(p) for p in paths]

  # Traverse sub-folders if required
  _walk = lambda p, t, r, ptn=None: walk(
    p, type_filter=t, pattern=ptn, re_pattern=re_pattern,
    ignored_patterns=ignored_patterns, return_basename=False, recursive=r,
    include_folder_name=include_folder_name)
  if recursive:
    for p in _walk(root_path, 'dir', False):
      # Walk into `p` first, if no targets found, skip
      sub_paths = _walk(p, type_filter, True, ptn=pattern)
      if len(sub_paths) == 0: continue

      if p in paths:
        paths.remove(p)
        paths.append(p)
      if p not in paths and include_folder_name:
        paths.append(p)
      paths.extend(sub_paths)
  return paths


def zip_dir(src_path, pattern, ignored_patterns=(), re_pattern=None,
            dst_path=None, overwrite=True, verbose=False,
            ignore_hidden_directories=True) -> str:
  """Create a ZIP file for all files with given pattern under `scr_path`

  :param src_path: source path
  :param pattern: patterns to be included using fnmatch
  :param ignored_patterns: patterns to be ignored using fnmatch
  :param re_pattern: patterns to be included using regular expression
  :param dst_path: directory to save .zip file
  :param overwrite: whether to overwrite if .zip file already exists
  :param verbose: whether or not to print processing details
  :param ignore_hidden_directories: whether to ignore hidden directories
  :return: path of the created .zip file
  """
  if verbose:
    console.show_status('Creating ZIP file for source directory:')
    console.split(color='yellow')
    tree(src_path, type_filter='file', pattern=pattern, re_pattern=re_pattern,
         ignored_patterns=ignored_patterns)
    console.split(color='yellow')

  # Check source path
  if not os.path.isdir(src_path):
    raise FileNotFoundError('!! Invalid directory `{}`.'.format(src_path))
  if src_path[-1] == '/': src_path = src_path[:-1]

  # Check destination path
  if dst_path is None: dst_path = src_path
  if os.path.isdir(dst_path):
    src_base_name = os.path.basename(src_path)
    dst_path = os.path.join(src_path, '{}.zip'.format(src_base_name))
  if not fnmatch(dst_path, '*.zip'):
    raise NameError('!! Illegal `dst_path`: {}'.format(dst_path))

  # Delete old zip file if exist
  if os.path.exists(src_path) and not overwrite:
    raise FileExistsError('!! ZIP file `{}` already exists.'.format(dst_path))

  # Create zip file
  file_paths = walk(src_path, type_filter='file', pattern=pattern,
                    re_pattern=re_pattern, ignored_patterns=ignored_patterns,
                    ignore_hidden_directories=ignore_hidden_directories,
                    recursive=True)
  with zipfile.ZipFile(dst_path, 'w', zipfile.ZIP_DEFLATED) as zipf:
    start = os.path.join(src_path, '..')
    for p in file_paths: zipf.write(p, os.path.relpath(p, start))

  # Finalize
  if verbose: console.show_status('ZIP file saved to `{}`.'.format(dst_path))
  return dst_path


def unzip_file(zip_path: str, dst_path: str, verbose=False):
  """Unzip a .zip file to `dst_path`

  :param zip_path: path of the .zip file to be unzipped
  :param dst_path: destination path
  :param verbose: whether to display processing details
  """
  if not zipfile.is_zipfile(zip_path):
    raise FileExistsError('!! Can not find ZIP file `{}`'.format(zip_path))

  with zipfile.ZipFile(zip_path, 'r') as zipf: zipf.extractall(dst_path)

  if verbose:
    console.show_status('`{}` unzipped to `{}`'.format(zip_path, dst_path))

  return dst_path


def tree(root_path: str, type_filter=None, pattern=None, ignored_patterns=(),
         re_pattern=None):
  """List a directory tree using walk. Example output:

  root_path
  |--foo.py
  |--sub-folder-1
  |--sub-folder-2
  |  |--sub-folder-2-1
  |     |--bar.py
  |- sub-folder-3

  :param root_path: directory to be traversed
  :param type_filter: specify what type of path should be returned, can be
                      (1) None (default), indicating return all paths
                      (2) `file` for filtering out file paths
                      (3) `folder` or `dir` for filtering out directories
  :param pattern: further filter out paths using fnmatch
  :param ignored_patterns: patterns to be ignored using fnmatch
  :param re_pattern: further filter out paths using regular expression
  """
  # Calculate root level
  if root_path[-1] == '/': root_path = root_path[:-1]
  level = lambda p: len(p.split('/'))
  root_level = level(root_path)
  get_prefix = lambda p: '|  ' * (level(p) - root_level - 1) + '|--'

  # Recursively get paths
  paths = walk(root_path=root_path, type_filter=type_filter, pattern=pattern,
               ignored_patterns=ignored_patterns, return_basename=False,
               re_pattern=re_pattern, recursive=True, include_folder_name=True)

  # Print tree
  console.write_line(console.fancify(root_path, 'bold'))
  for p in paths:
    prefix = console.fancify(get_prefix(p), 'white')
    name = os.path.basename(p)
    # Use bold font for directories
    if os.path.isdir(p): name = console.fancify(name, 'bold')
    console.write_line(prefix + name)


def synchronize(src_dir, dst_dir, pattern=None, ignored_patterns=(),
                re_pattern=None, ignore_hidden_directories=True, verbose=False):
  """Synchronize `dst_dir` with `src_dir`. This method was designed under the
  need of synchronizing the project directory on a GPU server with the local
  project directory which is hierarchical.

  :param src_dir: source directory
  :param dst_dir: destination directory
  :param pattern: patterns to be included using fnmatch
  :param ignored_patterns: patterns to be ignored using fnmatch
  :param re_pattern: patterns to be included using regular expression
  :param ignore_hidden_directories: whether to ignore hidden directories
  :param verbose: whether or not to print processing details
  """
  # Sanity check
  for path, name in zip((src_dir, dst_dir), ('Source', 'Destination')):
    if not os.path.exists(path): raise FileNotFoundError(
      '!! {} path `{}` not found.'.format(path, name))

  if os.path.basename(src_dir) != os.path.basename(dst_dir):
    raise AssertionError('!! Base name between source path `{}` and destination'
                         ' path `{}` should be the same.'.format(
      src_dir, dst_dir))

  # Create a .zip file from source path
  zip_path = zip_dir(src_dir, pattern, ignored_patterns, re_pattern,
                     ignore_hidden_directories=ignore_hidden_directories,
                     verbose=verbose)
  # Unzip
  unzip_file(zip_path, os.path.dirname(dst_dir), verbose=verbose)







# Copyright 2020 William Ro. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""This module defines basic class for build-functions"""

from ...spqr.censor import check_type
from .. import city


class Scroll(object):

  names = []

  def __init__(self, city_):
    self.city = check_type(city_, city.City)


  def __call__(self, *args, **kwargs):
    raise NotImplementedError


  def register(self, func_dict):
    check_type(func_dict, dict)
    assert len(self.names) > 0
    for name in self.names: func_dict[name] = self



# Copyright 2020 William Ro. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ------------------------------------------------------------------------------
# This file incorporates work covered by the following copyright and
# permission notice:
#
#   Copyright (c) 2008-2011 Volvox Development Team
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to
#   deal in the Software without restriction, including without limitation the
#   rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#   sell copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
#   IN THE SOFTWARE.
#
#   Author: Konstantin Lepa <konstantin.lepa@gmail.com>
# ====-==================================================================-======
"""Methods for printing colored symbols, may be with special attributes.

This module depends on termcolor package.
See https://pypi.org/project/termcolor/

Examples:
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from termcolor import colored, cprint

# The lines below will yield exactly the same outputs
cprint('Hello, World!', 'red', 'on_cyan')
write_line('Hello, World!', 'red', 'on_cyan')
write_line('#{Hello, World!}{red}{on_cyan}')

# The lines below will yield exactly the same outputs
print(colored('Hello', 'red') + ', ' + colored('World', 'green') + '!')
write_line('#{Hello}{red}, #{World}{green}!')
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
"""
import os
import re
import time
from sys import stdout


__all__ = ['parse', 'write_line', 'write', 'clear_line']

fancy_text = True


# region: Code from termcolor.py

ATTRIBUTES = dict(
        list(zip([
            'bold',
            'dark',
            '',
            'underline',
            'blink',
            '',
            'reverse',
            'concealed'
            ],
            list(range(1, 9))
            ))
        )
del ATTRIBUTES['']


HIGHLIGHTS = dict(
        list(zip([
            'on_grey',
            'on_red',
            'on_green',
            'on_yellow',
            'on_blue',
            'on_magenta',
            'on_cyan',
            'on_white'
            ],
            list(range(40, 48))
            ))
        )


COLORS = dict(
        list(zip([
            'grey',
            'red',
            'green',
            'yellow',
            'blue',
            'magenta',
            'cyan',
            'white',
            ],
            list(range(30, 38))
            ))
        )


RESET = '\033[0m'


def colored(text, color=None, on_color=None, attrs=None):
    """Colorize text.

    Available text colors:
        red, green, yellow, blue, magenta, cyan, white.

    Available text highlights:
        on_red, on_green, on_yellow, on_blue, on_magenta, on_cyan, on_white.

    Available attributes:
        bold, dark, underline, blink, reverse, concealed.

    Example:
        colored('Hello, World!', 'red', 'on_grey', ['blue', 'blink'])
        colored('Hello, World!', 'green')
    """
    if all([color is None, on_color is None, attrs is None]): return text
    if os.getenv('ANSI_COLORS_DISABLED') is None and fancy_text:
        fmt_str = '\033[%dm%s'
        if color is not None:
            text = fmt_str % (COLORS[color], text)

        if on_color is not None:
            text = fmt_str % (HIGHLIGHTS[on_color], text)

        if attrs is not None:
            for attr in attrs:
                text = fmt_str % (ATTRIBUTES[attr], text)

        text += RESET
    return text

# endregion: Code from termcolor.py

# region: Enumerates

class TextColors:
  GREY = 'grey'
  RED = 'red'
  GREEN = 'green'
  YELLOW = 'yellow'
  BLUE = 'blue'
  MAGENTA = 'magenta'
  CYAN = 'cyan'
  WHITE = 'white'

class TextHighlights:
  ON_GREY = 'on_grey'
  ON_RED = 'on_red'
  ON_GREEN = 'on_green'
  ON_YELLOW = 'on_yellow'
  ON_BLUE = 'on_blue'
  ON_MAGENTA = 'on_magenta'
  ON_CYAN = 'on_cyan'
  ON_WHITE = 'on_white'

class Attributes:
  BOLD = 'bold'
  DARK = 'dark'
  UNDERLINE = 'underline'
  BLINK = 'blink'
  REVERSE = 'reverse'
  CONCEALED = 'concealed'

# endregion: Enumerates

# region: Private Methods

def _render(text, *args):
  """Renders text using colored.

  Arguments in the back, if exist, will overwrite the previous ones,
  except for ATTRIBUTE arguments, which will be gathered to a list.

  :param text: text to be rendered
  :param args: arguments for rendering
  :return: rendered text and raw text
  """
  color, highlight, attributes = None, None, []
  for arg in args:
    if arg in COLORS.keys():
      color = arg
    elif arg in HIGHLIGHTS.keys():
      highlight = arg
    elif arg in ATTRIBUTES.keys() and arg not in attributes:
      attributes.append(arg)
  return colored(text, color, highlight, attributes), text

# endregion: Private Methods

# region: Public Methods

def parse(text, lead='#'):
  """Parse the given text according to the syntax.

  Pattern: lead + r'(\{.+?\})+'
  Here 'lead' is the leading character that can be specified.
  e.g., if 'lead' is set to '#', the pattern will be r'#(\{.+?\})+'

  Example:
  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  from termcolor import colored

  rendered, raw = parse('#{Hello}{red}, #{World}{green}!')
  assert raw == 'Hello, World!'
  # The lines below will yield exactly the same outputs
  print(rendered)
  print(colored('Hello', 'red') + ', ' + colored('World', 'green') + '!')
  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  :param text: content to be parsed
  :param lead: the leading character for parsing, '#' by default.
               BETTER NOT CHANGE !!
  :return: rendered line for and raw line, see example above
  """
  assert isinstance(text, str) and isinstance(lead, str)
  # Find units to be colored
  pattern = lead + r'(?:\{.+?\})+'
  units = re.findall(pattern, text)
  # Render each unit
  rendered_text, raw_text = text, text
  for unit in units:
    args = re.findall(r'\{(.+?)\}', unit)
    ren, raw = _render(*args)
    # Update rendered and raw text
    rendered_text = rendered_text.replace(unit, ren)
    raw_text = raw_text.replace(unit, raw)

  # Force to display plain text if fancy text is disabled
  if not fancy_text: rendered_text = raw_text
  return rendered_text, raw_text


def write_line(text, color=None, highlight=None, attributes=None, **kwargs):
  """Write a line with fancy style.

  During parsing, changing the leading character is not supported currently.

  :param text: text to be written
  :param color: should be in
        {red, green, yellow, blue, magenta, cyan, white}
  :param highlight: should be in
        {on_red, on_green, on_yellow, on_blue, on_magenta, on_cyan, on_white}
  :param attributes: should be a subset of
        {bold, dark, underline, blink, reverse, concealed}
  :param kwargs: additional keyword arguments for python print function
  :return: raw text
  """
  ren, raw = parse(text)
  print(colored(ren, color, highlight, attributes), **kwargs)
  return raw


def write(text, color=None, highlight=None, attributes=None):
  """Write the given text with fancy style.

  During parsing, changing the leading character is not supported currently.

  :param text: text to be written
  :param color: should be in
        {red, green, yellow, blue, magenta, cyan, white}
  :param highlight: should be in
        {on_red, on_green, on_yellow, on_blue, on_magenta, on_cyan, on_white}
  :param attributes: should be a subset of
        {bold, dark, underline, blink, reverse, concealed}
  """
  ren, _ = parse(text)
  stdout.write(colored(ren, color, highlight, attributes))
  stdout.flush()


def print_progress(index=None, total=None, start_time=None, progress=None,
                   bar_width=65):
  """Show progress bar.

  This method is inherited from tframe.utils.console.print_progress.
  The line which the cursor in terminal is positioned will be overwritten.

  :param index: positive scalar, indicating current working progress
  :param total: positive scalar, indicating the scale of total work
  :param start_time: if provided, ETA will be displayed to the right of
                      the progress bar
  :param progress: if provided, 'index' and 'total' will be ignored.
  :param bar_width: width of progress bar, 65 by default
  """
  # Calculate progress if not provided
  if progress is None:
    if index is None or total is None:
      raise ValueError(
        'index and total must be given if progress is not provided')
    progress = 1.0 * index / total
  progress = min(progress, 1.0)

  # Generate tail
  if start_time is not None:
    duration = time.time() - start_time
    eta = duration / max(progress, 1e-7) * (1 - progress)
    tail = "ETA: {:.0f}s".format(eta)
  else:
    tail = "{:.0f}%".format(100 * progress)

  # Generate progress bar
  left = int(progress * bar_width)
  right = bar_width - left
  mid = '=' if progress == 1 else '>'
  clear_line()
  stdout.write('[%s%s%s] %s' %
               ('=' * left, mid, ' ' * right, tail))
  stdout.flush()


def clear_line(text_width=79):
  """Clear a line in console."""
  stdout.write("\r{}\r".format(" " * text_width))
  stdout.flush()

# endregion: Public Methods


if __name__ == '__main__':
  # The lines below will yield exactly the same outputs
  print(colored('Hello, World!', 'red', 'on_cyan'))
  write_line('Hello, World!', 'red', 'on_cyan')
  write_line('#{Hello, World!}{red}{on_cyan}')

  # The lines below will yield exactly the same outputs
  print(colored('Hello', 'red') + ', ' + colored('World', 'green') + '!')
  write_line('#{Hello}{red}, #{World}{green}!')
  write('#{Hello}{red}, #{World}{green}!')






from roma import console
from setuptools import setup, find_packages


# Specify version
VERSION = '1.0.0.dev6'


# Preprocess
console.show_status('Running setup.py for roma-v' + VERSION + ' ...')
console.split('-')


# Run setup
def readme():
  with open('README.md', 'r') as f:
    return f.read()

# Submodules will be included as package data, specified in MANIFEST.in
setup(
  name='py-roma',
  packages=find_packages(),
  include_package_data=True,
  version=VERSION,
  description='A python development kit containing handy packages.',
  long_description=readme(),
  long_description_content_type='text/markdown',
  author='William Ro',
  author_email='willi4m@zju.edu.cn',
  url='https://github.com/WilliamRo/roma',
  download_url='https://github.com/WilliamRo/roma/tarball/v' + VERSION,
  license='Apache-2.0',
  keywords=['utilities', 'console', 'checker'],
  classifiers=[
    "Development Status :: 2 - Pre-Alpha",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: Apache Software License",
    "Natural Language :: English",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 3",
    "Topic :: Utilities",
  ],
)



# Copyright 2020 William Ro. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ====-==================================================================-======
"""TODO: add description here (v1.0.0.dev5)"""

from .censor import check_type
from collections import OrderedDict

import re


class Arguments(object):
  """TODO: description (v1.0.0.dev5)"""

  def __init__(self, name: str, *args, **kwargs):
    self.func_name = check_type(name, str)
    self.arg_list = args
    self.arg_dict = OrderedDict(kwargs)


  def __str__(self):
    return '{}({})'.format(self.func_name, ', '.join(
      list(self.arg_list) +
      ['{}={}'.format(k, v) for k, v in self.arg_dict.items()]))


  @staticmethod
  def parse(arg_string: str):
    """Parse arg_string and return an instance of Arguments.

    The format of arg_string should be:
    'entry_name arg1 ... argN kwarg1=v1 ... kwargM=vM'

    The best usage of this method is parsing the arguments for a specific
    python method. For example, given a method

        def add(x, y, shift=0):
          return x + y + shift

    and a command line argument string

        arg_string = 'add 12 19 shift=316'

    We have

        a = Arguments.parse(arg_string)
        assert add(*a.arg_list, **a.arg_dict) == 347

    :param arg_string: argument string in a given format above
    :return: an instance of Arguments
    """
    check_type(arg_string, str)
    pieces = arg_string.split(' ')
    assert len(pieces) > 0

    # (1) func_name
    func_name = pieces.pop(0)
    r = re.match(r'^[a-z_]+\w+$', func_name)
    if r is None: raise ValueError(
      "function name '{}' is illegal.".format(func_name))

    # (2) arg_list
    args = []
    while len(pieces) > 0 and '=' not in pieces[0]:
      args.append(pieces.pop(0))

    # (3) arg_dict
    kwargs = OrderedDict()
    while len(pieces) > 0:
      piece = pieces.pop(0)
      r = re.match(r'(^[a-zA-Z]+)=(.+)', piece)
      if r is None or len(r.groups()) != 2: raise ValueError(
        "Illegal string for kwargs: " + piece)
      k, v =  r.groups()
      kwargs[k] = v

    return Arguments(func_name, *args, **kwargs)




# Copyright 2020 William Ro. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ====-==============================================================-==========
"""City means node"""

from ..spqr import censor
from ..console.console import console
from . import hall

import socket


class City(object):
  """

  """

  def __init__(self, ip='', port=0, name='Roma', backlog=5, libs=None):
    """"""
    # Public attributes
    self.name = censor.check_type(name, str)

    # : Private attributes
    # Check libs and instantiate city hall
    if libs is None: libs = []
    elif isinstance(libs, str): libs = [libs]
    self._hall = hall.Hall(self, *censor.check_type(libs, inner_type=str))

    self._backlog = censor.check_type(backlog, int)

    # initiate a socket
    self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    # bind socket
    self.socket.bind((ip, port))
    self._ip, self._port = self.socket.getsockname()
    console.show_status('City {} (located in {}) has been created.'.format(
      self.name, console.fancify(self.location_str, 'blue', 'underline')))


  # region: Properties

  @property
  def location(self):
    return self._ip, self._port

  @property
  def location_str(self):
    return '{}:{}'.format(self._ip, self._port)

  # endregion: Properties

  # region: Private Methods

  def _host(self):
    console.show_status('{} is hosting ...'.format(self.name))

  # endregion: Private Methods

  # region: Public Methods

  def enter_hall(self):
    """Enter the hall of this city"""
    console.show_status('Entered city hall.')
    while True:
      console.write('<= ', color='green')
      cmd = input()
      try:
        if self._hall.handle(cmd): return
      except Exception as e:
        console.show_status(
          'Fail to execute command \'{}\'. Error message:'.format(cmd),
          prompt='!!', color='red')
        console.supplement(e, color='red')

  # endregion: Public Methods

"""

    # Put this city on listen mode
    self._socket.listen(backlog)
"""



